/*
 * kgraft_patch_bsc1039496
 *
 * Fix for SLE specific regression introduced by ("mm: enlarge stack
 * guard gap").
 *
 *  Offending commits:
 *  SLE12-SP2 commit: d016488325eaa0a060aa72d24cd9c07e904b6f4c
 *  SLE12(-SP1) commit: d14041c8b67b14a658572849b0ec7a9cb9c0acd3
 *
 *  The resolution for the SLE12* kernels has been to replace the
 *  offending commit with the upstream solution.
 *
 *  Thus, this livepatch is based on
 *  SLE11-SP4 commit cf4d269b082a02f96d205db88c6622ceb5908815
 *
 *  Copyright (c) 2017 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kconfig.h>
#include <linux/kallsyms.h>
#include <linux/mm.h>


#if defined(CONFIG_STACK_GROWSUP)
#error "KGR patch supports only CONFIG_STACK_GROWSUP=n"
#endif

#ifndef CONFIG_MMU
#error "KGR patch supports only CONFIG_MMU=y"
#endif


static unsigned long *kgr_stack_guard_gap;
static unsigned long *kgr_mmap_min_addr;

static struct {
       char *name;
       char **addr;
} kgr_funcs[] = {
	{ "stack_guard_gap", (void *)&kgr_stack_guard_gap },
	{ "mmap_min_addr", (void *)&kgr_mmap_min_addr },
};


unsigned long kgr_expandable_stack_area(struct vm_area_struct *vma,
		unsigned long address, unsigned long *gap)
{
	struct vm_area_struct *prev = vma->vm_prev;
	unsigned long guard_gap = *kgr_stack_guard_gap;
	unsigned long guard_addr;

	address &= PAGE_MASK;
	if (!prev)
		goto out;

	/*
	 * Is there a mapping abutting this one below?
	 *
	 * That's only ok if it's the same stack mapping
	 * that has gotten split or there is sufficient gap
	 * between mappings
	 */
	/*
	 * Fix SLE regression
	 *  -1 line, +1 line
	 */
	if ((prev->vm_flags & VM_GROWSDOWN) || !(prev->vm_flags & (VM_WRITE|VM_READ))) {
		guard_gap = min(guard_gap, address - prev->vm_end);
		goto out;
	}

	if (address - prev->vm_end < guard_gap)
		return -ENOMEM;

out:
	/* make sure we won't underflow */
	if (address < *kgr_mmap_min_addr)
		return -ENOMEM;
	if (address - *kgr_mmap_min_addr < guard_gap)
		guard_gap = address - *kgr_mmap_min_addr;

	guard_addr = address - guard_gap;
	*gap = guard_gap;

	return guard_addr;
}

static int kgr_patch_bsc1039496_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("kgraft-patch: symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}

		*(kgr_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

int kgr_patch_bsc1039496_init(void)
{
	return kgr_patch_bsc1039496_kallsyms();
}
